"""Set metadata for chat-downloader"""

__title__ = 'chat-downloader'
__program__ = 'chat_downloader'
__summary__ = 'A simple tool used to retrieve chat messages from livestreams, videos, clips and past broadcasts. No authentication needed!'
__author__ = 'xenova'
__email__ = 'admin@xenova.com'
__copyright__ = '2020-2022 xenova'
__url__ = 'https://github.com/xenova/chat-downloader'
__version__ = '0.2.0'
