"""Set metadata for chat-analyzer"""

__title__ = 'chat-analyzer'
__program__ = 'chat_analyzer'
__summary__ = 'A tool used to process and analyze chat data from past live streams, providing summarized information about chat activity over the stream\'s lifetime.'
__author__ = 'David Fryd'
__email__ = 'info@chatanalyze.com'
__copyright__ = '2022 David Fryd'
__license__ = 'MIT'
__url__ = 'https://github.com/David-Fryd/chat-analyzer'
__website__ = 'https://chatanalyze.com'
__version__ = '1.0.0b5'
